/* $Id: mailerlib.h,v 2.16 2017/12/03 12:28:58 hurtta Exp $ */

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.16 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> (was hurtta+elm@posti.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

/* init.c */

extern void init_mailerlib P_((void));

extern PATH valid_domains;

/* mailer.c */

extern int mailerfunc P_((char **value, int enter,
			  int lineno, const char *filename));

typedef void sending_message_func P_((int background));

extern struct mailer_info  *get_mailer_info P_((void));
extern void free_mailer_info P_((struct mailer_info **S));

enum MI_query { MI_HAVE_8BITMIME = 1, MI_HAVE_BINARYMIME = 2,
		MI_HAVE_DSN      = 3, MI_DONT_ADD_FROM  = 4,
		MI_USE_DOMAIN    = 5, MI_HAVE_SENDING_DISPLAY = 6
};
	
		    
extern int query_mailer_info P_((struct mailer_info  *I,
				 enum MI_query query));

/* -1  if connection lost
    0  if OK
    1  if mailer reinitialized and  query_mailer_info()
          need to be called again
*/
extern int mailer_restarted P_((struct mailer_info  *I));

/* Should return 1 if verify succees and fields of result to be filled 
   -- caller should free fields of result
   Returns 0 on failure
*/

enum mailer_errcode {
    MAILER_OK = 0,                  /* Address OK */
    MAILER_NOT_AVAIL,               /* Address verify not available */
    MAILER_NOT_EXIST                /* Address not exists */
};

struct addr_verify_result {   /* OLD struct addr_item */
    char *addr;
    struct string *fullname;
    struct string *comment;
};

int verify_mailer_addr P_((struct mailer_info  *I,
			   const char *text,
			   struct addr_verify_result *result,
			   enum mailer_errcode *errcode));
int verify_mailer_domaddr P_((struct mailer_info  *I,
			   const char *text,
			   struct addr_verify_result *result,
			   enum mailer_errcode *errcode));





/* Return NULL if no editor (and no value either) */
extern struct mailer_env_from * mailer_get_env_from P_((struct mailer_info  *I));
extern void mailer_free_env_from P_((struct mailer_env_from **X));
extern int env_from_changed P_((struct mailer_env_from *X));

/* return temporary pointer  value -- do not free 
   return NULL if currently no value set (use implicit default)
  */
extern const char * mailer_env_from_value P_((struct mailer_env_from *X,
					    int *can_edit));

extern void mailer_env_from_change P_((struct mailer_env_from *X,
				       const char * value));

extern struct mail_send_state  *mailer_init P_((char **addr_args,
						int dsn,
						int verbose,
						struct mailer_info *info,
						struct mailer_env_from *env_from));


/* Note that also mail_backend3 does freeing ... */
extern void free_mail_send_state P_((struct mail_send_state **S));


extern struct out_state *get_mail_outfd P_((struct mail_send_state  *X));


extern struct mailer_cancel_ticket * new_mailer_cancel_ticket P_((void));
extern void free_mailer_cancel_ticket P_((struct mailer_cancel_ticket **
					  ticket));

extern enum msd_action { 
    msd_act_EOF        = -1  /* EOF seen on user input       */,
    msd_act_SIG        = -2  /* Ctrl-C or OE_SIG_CHAR        */,
    msd_act_need_restart = -3  /* mailer_restarted() need to be called 
				  (cancel copy) then send again           */,

    msd_act_default  = 0  /* default action or not defined or OK */,
    msd_act_cancel        /* forget/cancel or do not send  (cancel copy)  */,
    msd_act_edit          /* 'edit' again                  (cancel copy)  */,
    msd_act_send          /* 'send' or continue anyway    */
} have_msd_action P_((struct mailer_cancel_ticket *ticket));

extern const char * have_cancel_filename  P_((struct mailer_cancel_ticket *
					      ticket));
extern FILE * have_cancel_FILE P_((struct mailer_cancel_ticket *
					      ticket));
/* free_mailer_cancel_ticket closes / unlinks this file (if exists) */



/* mail_backend3 may free (*mail_fd) pointer ... */
extern int mail_backend3 P_((struct mail_send_state **mail_fd,
			     end_handler *func,
			     int encoding_top,
			     char * title,
			     sending_message_func *sm,
			     int *exit_status  /* may be NULL */,

			     struct mailer_cancel_ticket *
			     ticket  /* may be NULL */,
			     long message_size));


/* Return mailer path or type ... */
extern const char *get_mailer_path P_((struct mail_send_state *mail_fd));

void init_default_mailer P_((void));

extern char * kludge_addr P_((char **addr));

/* mail sending (recipient) display */

#if ANSI_C
struct sending_display;
#endif

/* callbacks */

typedef void msd_free_sending_display_cb P_((struct sending_display **sd));


typedef void msd_inc_sending_display_cb P_((struct sending_display *sd));


enum msd_addr_status { msd_addr_tmpfail  = -1,
		       msd_addr_fail     = -2,
		       msd_addrstat_none = 0, /* status not yet available */
		       msd_sending_addr  = 1,
		       msd_addr_ok       = 2
};


/* idx == -1  is sender (MAIL FROM) 
   idx == 0 .. (count of addr_args from mailer_init) is recipient (RCPT TO)
*/
typedef void msd_update_addr_cb P_((struct sending_display *sd,
				    int                    idx,
				    struct string        * addr,
				    enum msd_addr_status   status,
				    struct string        * message /* server */
				    ));
	

enum msd_can_continue { msd_mail_failed = 0, msd_address_problem };


/* Return 0 (msd_act_default) on timeout, wait ~second for user input 
   idx == -1  is sender (MAIL FROM) 
   idx == 0 .. (count of addr_args from mailer_init) is recipient (RCPT TO)

   block (wait user) if not can_continue

   Called if at one address fails 
   AND struct mailer_cancel_ticket is available 
*/

typedef enum msd_action msd_check_addr_cb P_((struct sending_display *sd,
					      int                    idx,
					      struct string        * addr,
					      enum msd_addr_status   status,
					      struct string  * message 
					      /* server */,
					      enum msd_can_continue 
					      can_continue));
			      
enum msd_data_status { msd_mail_tmpfail = -1,
		       msd_mail_fail    = -2,
		       msd_mail_backgrounded = -3,  /* Can not pass ticket */
		       msd_calling_mailer = -4, /* calling mailer      */
		       msd_datastat_none  = 0,  /* status not yet available */
		       msd_sending_data = 1,
		       msd_data_ok      = 2,    /* DATA command accepted */
		       msd_mail_sent    = 3     /* Final dot OK */
};

typedef void msd_update_data_cb P_((struct sending_display * sd,
				    enum msd_data_status     status,
				    struct string    * message /* server */));


/* Called if at least one address or DATA fails 
   AND struct mailer_cancel_ticket is available 

   Not called if  msd_check_addr_cb already returns msd_action
*/
typedef enum msd_action msd_check_fail_cb P_((struct sending_display *sd,
					      enum msd_data_status   status,
					      struct string  * message /* server */,
					      enum msd_can_continue 
					      can_continue));

enum msd_connection_info { msd_server = 0,
			   msd_verified_as    /* if certificate ok */,

			   NUM_msd_connection_info
};

typedef void msd_update_con_cb  P_((struct sending_display *sd,
				    enum msd_connection_info idx,
				    struct string *text));

/* Called when mail is too big */
typedef enum msd_action msd_check_size_cb P_((struct sending_display *sd,
					      long message_size,
					      long size_limit));


enum msd_size_info   { msd_message_size = 0,
		       msd_size_limit,
		       NUM_msd_size_info };

typedef void msd_update_size_cb P_((struct sending_display *sd,
				    enum msd_size_info idx,
				    long size));

/* 0 .. not supported 
   1 .. supported 

   calls msd_inc_sending_display_cb
   if take reference from struct sending_display *sd
*/

extern int mailer_set_sending_display 
                  P_((struct mail_send_state      * S,
		      struct sending_display      * sd /* opaque */,
		      msd_free_sending_display_cb * free_cb,
		      msd_inc_sending_display_cb  * inc_cb,
		      msd_update_addr_cb          * addr_upd_cb,
		      msd_check_addr_cb           * addr_check_cb,
		      msd_update_data_cb          * data_upd_cb,
		      msd_check_fail_cb           * check_fail_cb,
		      msd_update_con_cb           * con_upd_cb,
		      msd_check_size_cb           * check_size_cb,
		      msd_update_size_cb          * size_upd_cb
		      ));

/* outheaders.c */

struct mailing_headers;
extern void zero_mailing_headers P_((struct mailing_headers *hdrs));
extern void free_mailing_headers P_((struct mailing_headers *hdrs));

/* Result is malloced */
extern char * from_addr_literal1 P_((int add_dom));
extern char * from_addr_literal P_((struct mailer_info *mailer_info));

struct expanded_address;

extern struct string * username_to_surface_addr P_((void));
extern void make_from_addr P_((struct expanded_address *expanded,
			      struct mailer_info *mailer_info));

/* Returns domain name if matches to valid-domains */

extern const char * is_whitelisted_valid_domain P_((const char * mail_domain_name));

/* shared_mailer.c */

#ifdef USE_DLOPEN
extern SHAREDLIB use_shared_mailer;
#endif


extern struct mailer_config  * loc_mailer_type P_((char *f));

extern enum verify_domain_result { 
    verify_domain_not_available = -3 /* no shared library or routine */,
	verify_domain_fallback  = -2 /* call default routine instead */,
	verify_domain_failure   = -1   /* can't verify */,
	verify_domain_not_found  = 0   /* domain not found */,
	verify_domain_ok         = 1   /* domain found */,
	verify_domain_no_mail    = 2   /* domain accept no mail */,
	verify_domain_bad_syntax = 3   /* domain have bad syntax       */
	} shared_verify_domain P_((const char *domain,char **rewrite,
				   const char *tag, int translated_value,
				   const struct dt_enum_shared * is_shared_value,
				   const struct schedule_timelimit * now,
				   struct schedule_timelimit       * valid_until,
				   struct cancel_data             ** cancel_p
				   /* May be NULL, Used if dns lookup was cancelable */

				   ));

/* mime_encode.c */

extern void base64_encode P_((FILE *srcfp,
			      struct out_state *mailer,
			      int istext,
			      int top_encoding,
			      int source_is_binary));

extern void line_quoted_printable_encode P_((char *input, 
					     struct out_state *mailer, int len, 
					     int istext,
					     int top_encoding,
					     int source_is_binary));


extern void quoted_printable_encode  P_((FILE *srcfp, struct out_state *mailer, 
					 int istext,
					 int top_encoding,
					 int source_is_binary));

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
